CREATE OR REPLACE PACKAGE BODY
  -- ------------------------------------------------------------------------
  --     COPYRIGHT IBM 2003-2007
  -- @(#) BH_Update.sql 1.91.0 08/01/24 16:34:30
  -- ------------------------------------------------------------------------
  BH_Update IS
   
   -- SLEEP_TIME_INITIAL = #seconds to initially sleep between retries
   -- SLEEP_TIME_MULT = multiplier between retries in exp backoff
   -- MAX_RETRY = maximum number of retries
   -- With this configuration the maximum total time slept will be 126s.
   SLEEP_TIME_INITIAL CONSTANT INTEGER := 1;
   SLEEP_TIME_MULT    CONSTANT INTEGER := 2;
   MAX_RETRY          CONSTANT INTEGER := 4;
   
   g_task_lock_name  VARCHAR2(30) := 'busyhour-get-next-task-lock';

   gJobId  INTEGER;
   gTaskId INTEGER;   
   gOverRide VARCHAR2(2);
   gErrorString    VARCHAR2(4000) := 'Failed to store:';
   gErrorStringSet BOOLEAN := false;

   -- tunable parameters
   g_sort_area_size INTEGER := 10*1024*1024; -- sort_area_size for traffic queries
   g_sort_area_retained_size INTEGER := 10*1024*1024; -- ditto
   g_hash_area_size INTEGER := 2*g_sort_area_size;
   
   
   
   -- fiftnines is 59m59s (1 hour minus 1 second)
   -- We're going to use this to add an "hour" to some values
   -- so for code clarity, calculate it once...
   g_fiftynines CONSTANT FLOAT := (1/24) - 1/(24*60*60);  

   
   gFirstAvail  date := null;
   gLastAvail   date := null;

   g_OPS_ABORT            CONSTANT INTEGER := 1;        -- We have been interrupted, ABORT
   g_OPS_READY            CONSTANT INTEGER := 2;        -- Waiting for a busyhour_task
   g_OPS_ACTIVE           CONSTANT INTEGER := 3;        -- busyhour_task is processing
   g_OPS_SUCCESS          CONSTANT INTEGER := 4;        -- Success
   g_OPS_FAILED           CONSTANT INTEGER := 5;        -- Failed

   g_poll_interval     INTEGER := 10;               -- poll interval in seconds
   g_retry_interval    INTEGER := 5;                -- poll interval in seconds
   g_min_bh_sessions   INTEGER := 1;                -- This is the MIN number of sessions
   g_bh_sessions       INTEGER := 1;                -- This is the config number of sessions

   g_slices            INTEGER := 1;                -- Will default to g_bh_sessions after parameters are loaded
   g_slicethreshold    INTEGER := 0;                -- No slices is the default

   -- DBMS_Application_Info string
   g_module_name   CONSTANT VARCHAR2(30) := 'BusyHour';

   -- Global to hold the session ids
   session_id_tab bh_session_id_tabletype := bh_session_id_tabletype();


   PROCEDURE dump_options( job_id_in IN INTEGER );

   FUNCTION slice_msg(start_id IN INTEGER, end_id IN INTEGER ) RETURN VARCHAR2;

   PROCEDURE UpdateRGStored(task_id_in INTEGER,
			    bh_def PM_BusyHour%ROWTYPE,
			    startTime DATE,
			    endTime DATE);
   
   
   
   PROCEDURE PopulateCompleteTree(job_id_in INTEGER,
				  task_id_in INTEGER,
				 week_in DATE);
   
   PROCEDURE SummarizeAllTables(
				job_id_in INTEGER,
				task_id_in INTEGER,
				week_in DATE);
   
--   FUNCTION SummarizeTable( task_desc BHtaskRec ) RETURN BOOLEAN;
   
   PROCEDURE CleanOldBH(job_id_in IN INTEGER,
			task_id_in IN INTEGER,			
			clean_table IN VARCHAR2,
			table_entity IN INTEGER,
			startTime IN DATE,
			endTime IN DATE,
                        start_id IN INTEGER,
                        end_id IN INTEGER);
   
   PROCEDURE RGLocking(task_id_in INTEGER, start_or_stop VARCHAR2);
   
   PROCEDURE update_task(task_id_in INTEGER, message VARCHAR2);
   PROCEDURE add_to_error_string(table_name VARCHAR2);

   PROCEDURE clean_temp_tables(task_id_in INTEGER);
   
   FUNCTION job_interrupted( job_id_in IN INTEGER ) RETURN BOOLEAN;

   PROCEDURE update_bh_task( p_jobid IN INTEGER, 
                             p_taskid IN INTEGER, 
                             p_code IN INTEGER,
                             p_reset IN BOOLEAN DEFAULT FALSE );

   PROCEDURE update_ready_count( p_jobid IN INTEGER );

   PROCEDURE create_bh_sessions( p_jobid IN INTEGER, p_bh_sessions IN INTEGER );

   PROCEDURE abort_tasks( p_jobid IN INTEGER );

   FUNCTION all_sessions_complete RETURN BOOLEAN;

   FUNCTION all_tasks_success( p_jobid IN INTEGER ) RETURN BOOLEAN;

   FUNCTION waiting_tasks( p_jobid IN INTEGER ) RETURN BOOLEAN;

   PROCEDURE get_next_task( p_jobid IN INTEGER, p_task IN OUT NOCOPY BHtaskRec );

   FUNCTION task_poll( jobidparam IN INTEGER) RETURN BOOLEAN;

   PROCEDURE add_bh_task( p_jobid IN INTEGER, p_task_id IN INTEGER, p_task IN BHtaskRec );


   -- -----------------------------------------------------------------------'
   -- -----------------------------------------------------------------------'
   FUNCTION get_tunable( param_name VARCHAR2
                                           , param_default INTEGER)
                RETURN INTEGER IS
     parameter_val INTEGER;
   BEGIN
     parameter_val := Wm_Sysval.getinteger(param_name);
         RETURN NVL(parameter_val, param_default);
   EXCEPTION
         WHEN OTHERS THEN
           RETURN param_default;
   END get_tunable;
      
   -- -----------------------------------------------------------------------
   PROCEDURE init_tunable_parameters IS
    vprc VARCHAR2(30) := 'Init_Tunable_Parameters';
    l_sort_area_size INTEGER;
    l_sort_area_retained_size INTEGER;
   BEGIN
         -- The old style was to resuse the TrafficSortAreaSize but we need
         -- better control. So we introduce a new parameter and fall back if it's 0

         l_sort_area_size := get_tunable('BHSortAreaSize', 0);
         IF ( l_sort_area_size > 0 ) THEN
           -- found the new system value
           g_sort_area_size := l_sort_area_size;
         ELSE
           -- fallback to the old value
           g_sort_area_size := get_tunable('TrafficSortAreaSize', g_sort_area_size);
         END IF;

         l_sort_area_retained_size := get_tunable('BHSortAreaRetainedSize', 0);
         IF ( l_sort_area_retained_size > 0 ) THEN
           -- found the new system value
           g_sort_area_retained_size := l_sort_area_retained_size;
         ELSE
           -- fallback to the old value
           g_sort_area_retained_size := get_tunable('TrafficSortAreaRetainedSize', g_sort_area_retained_size);
         END IF;

         -- Special option just for BH
         g_hash_area_size := get_tunable('BHHashAreaSize', g_hash_area_size);

         -- Parameters for busyhour_tasks, we always need at least g_min_bh_sessions
         g_bh_sessions := greatest( get_tunable( 'BHSessions', g_bh_sessions ), g_min_bh_sessions ); 
         g_poll_interval := get_tunable( 'BHPollInterval', g_poll_interval );
         g_retry_interval := get_tunable( 'BHRetryInterval', g_retry_interval );
  
         g_slicethreshold := get_tunable( 'BHSliceThreshold', 0 );
         g_slices := get_tunable( 'BHSlices', g_bh_sessions );

   END init_tunable_parameters;

   -- -----------------------------------------------------------------------
   -- Do nothing FOR TESTING ONLY
   -- -----------------------------------------------------------------------
   PROCEDURE RunUpdateEmpty(task_id_in INTEGER) IS
   BEGIN
      AlterTunables;
   END;
   
   -- -----------------------------------------------------------------------
   -- RunUpdateLeave - Run the update as normal but don't clean out
   -- rg_busyhour_tmp       FOR TESTING ONLY!!!!
   -- -----------------------------------------------------------------------
   PROCEDURE RunUpdateLeave(task_id_in INTEGER) IS
   BEGIN
      RunUpdate(task_id_in, false);
   END;
   
   -- -----------------------------------------------------------------------
   -- RunUpdate - Store all of the busyhours found in rg_busyhour_tmp
   --     BE VERY CAREFULL CALLING OTHER PROCEEDURES!!!!
   --     We depend on rg_busyhour_tmp to be populated and it's a temporary
   --     table with ON COMMIT DELETE set.
   -- -----------------------------------------------------------------------
   PROCEDURE RunUpdate(task_id_in INTEGER, del BOOLEAN DEFAULT true) IS
      bh_def     PM_BusyHour%ROWTYPE;
      i_message VARCHAR2(1000) := NULL;
      startTime  DATE;
      endTime    DATE;
      i_FirstAvail DATE := NULL;
      i_LastAvail DATE := NULL;
      i_start_date DATE := NULL;
      i_end_date DATE := NULL;
      startTime_char VARCHAR2(20) := NULL;
      endTime_char VARCHAR2(20) :=  NULL;
      endCalc    DATE;
      update_first_avail BOOLEAN := true;
      update_last_avail BOOLEAN := true;
      maxLastAvail  Date;

   BEGIN
      
      SELECT bh.* INTO bh_def
	FROM PM_BusyHour bh, BH_ReportSpec rs
	WHERE rs.Id = task_id_in
	AND rs.BusyHourId = bh.id;
      
      
      SELECT max(tstamp), min(tstamp) INTO endTime, startTime
	FROM rg_busyhour_tmp;
      
      gTaskId := task_id_in;
      
      SELECT job_id INTO gJobId
	FROM active_task
	WHERE task_id = task_id_in;
            
      startTime := wm_date.trunc_to_last_period(startTime, bh_def.span);
      endTime := wm_date.trunc_to_last_period(endTime, bh_def.span);
      endCalc := wm_date.add_periods(endTime, bh_def.span) - 1/(60*60*24);
      
      -- keep the real firstAvail and lastAvail
      gFirstAvail := startTime;
      gLastAvail := endTime;

      startTime_char :=  to_char(startTime,'yyyy-mm-dd');
      endTime_char := to_char(endTime,'yyyy-mm-dd');
      i_message := ' Processing busy hour ' || bh_def.name
                || ' ' || startTime_char || ' - > '
                || endTime_char;
                  

      AlterTunables;
      UpdateRGStored(task_id_in, bh_def, startTime, endCalc);
      
      --
      -- 
      -- 
      commit;
      --
      -- RG_BUSYHOUR_TMP IS **GONE** It's ON COMMIT DELETE for right now.
      --
      update_task(task_id_in,i_message);
            
   END RunUpdate;
   
   
   -- -----------------------------------------------------------------------
   -- GlobalUpdate - Main COPY routine
   -- -----------------------------------------------------------------------
   PROCEDURE GlobalUpdate(task_id_in INTEGER) IS
      l_job_id INTEGER;
      l_task_id INTEGER;
      l_offset INTEGER;
      l_override VARCHAR2(2);
      TYPE week_tab_type IS TABLE OF DATE;
      week_tab week_tab_type;

      -- Track the weeks
      l_number_of_weeks INTEGER := 0;

      -- Status string
      l_msg VARCHAR2(64);

      l_start_date DATE;
      l_end_date DATE;
   BEGIN 
      
      AlterTunables;

      SELECT job_id INTO l_job_id FROM active_task WHERE task_id = task_id_in;

    BEGIN
      SELECT nvl(override,'N') into l_override 
      from active_job aj,bh_parameter bp 
      where job_id = l_job_id and
        sched_fk = bp.id;
      EXCEPTION
	WHEN NO_DATA_FOUND then
	  dbms_output.put_line('No data found for the override flag for this job in global update.');
	  l_override := 'N';
    END;

      gOverRide := l_override;
      l_task_id := Status.generic_start_task( l_job_id, NULL);
      Status.generic_end_task( l_job_id, l_task_id,
         'OVERRIDE flag: = ' || gOverRide );


      dump_options( l_job_id );
      
      SELECT wm_date.trunc_to_last_period(r.tstamp, 'W') wk BULK COLLECT INTO week_tab
	  FROM bh_refresh_sbh r, active_task tsk
	  WHERE tsk.job_id = l_job_id
	  AND tsk.task_id = r.task_id
	  GROUP BY wm_date.trunc_to_last_period(r.tstamp, 'W')
	ORDER BY wm_date.trunc_to_last_period(r.tstamp, 'W') DESC;
      
      -- How many weeks?
      l_number_of_weeks := week_tab.COUNT;

      dbms_application_info.set_module( g_module_name, NULL );

      FOR l_offset IN week_tab.first..week_tab.last LOOP 

         -- Message ' [1/2 :: 20070101] '
         l_msg := ' [' 
            || to_char(l_offset)
            || '/'
            || to_char(l_number_of_weeks)
            || ' :: '
            || to_char(week_tab(l_offset), 'YYYYMMDD')
            || '] ';
 
         -- Prep the BH_Driver_tmp table
         dbms_application_info.set_client_info( 'PopulateCompleteTree tree' || l_msg );
	 PopulateCompleteTree(l_job_id, task_id_in, week_tab(l_offset));

	 -- Warning: PopulateCompleteTree has a COMMIT
         -- Prep for sessions
         dbms_application_info.set_client_info( 'Create_BH_Sessions' || l_msg );
         create_bh_sessions( l_job_Id, g_bh_sessions );

         dbms_application_info.set_client_info( 'SummarizeAllTables' || l_msg );
	 SummarizeAllTables(l_job_id, task_id_in, week_tab(l_offset));


         -- Have we been interruped ... already?
         IF ( job_interrupted( l_job_id ) ) THEN

            -- Mark all READY TASKS as ABORT so that we starve the sessions
            abort_tasks( l_job_id );

            Status.update_job_status(l_job_id, Sched_General.tm_message,
               'BusyHour failed - interrupted');
         END IF;


         -- Start the sessions ...
         COMMIT;

         -- Poll until all sessions are complete ... we may go again
         -- We must do them in pieces because we need the bh_driver_tmp repopulated
         dbms_application_info.set_client_info( 'Polling...' || l_msg );

         IF ( task_poll( l_job_id ) ) THEN


            SELECT min(tstamp), max(tstamp)+g_fiftynines INTO l_start_date, l_end_date
               FROM bh_refresh_sbh r, active_task tsk
               WHERE tsk.job_id = l_job_id
               AND tsk.task_id = r.task_id
               AND r.tstamp >= week_tab(l_offset)
               AND r.tstamp < week_tab(l_offset) + 7;
 
            UPDATE pm_busyhour pmbh
              SET (firstavail, lastavail) = 
                 (SELECT least(nvl(min(trunc(sbh.tstamp)),pmbh.firstavail),
                  nvl(pmbh.firstavail,min(trunc(sbh.tstamp)))),
                  greatest(nvl(max(trunc(sbh.tstamp)),pmbh.lastavail),
                  nvl(pmbh.lastavail,max(trunc(sbh.tstamp))))
                 FROM bh_refresh_sbh sbh,
                 active_task tsk
                 WHERE tsk.job_id = l_job_id
                 AND tsk.task_id = sbh.task_id
                 AND sbh.summary_id = pmbh.id
                 AND sbh.tstamp BETWEEN l_start_date AND l_end_date)
              WHERE field_id IS NOT NULL;
         ELSE
           -- ???
           NULL;
         END IF;

      END LOOP;
      
      dbms_application_info.set_client_info( 'Clean_Temp_Tables' || l_msg );
      clean_temp_tables(task_id_in);

      EXCEPTION
         WHEN OTHERS THEN
            -- It looks like the main routine checks the status of the job
            RAISE;
      
   END GlobalUpdate;
   
   
   -- -----------------------------------------------------------------------
   -- UpdateRGStored - Updates the stored info in RG_STORED_DBH.  Includes
   -- retry logic similar to SummarizeSingleHour.
   -- -----------------------------------------------------------------------
   PROCEDURE UpdateRGStored(task_id_in INTEGER,
			    bh_def PM_BusyHour%ROWTYPE,
			    startTime DATE,
			    endTime DATE) IS
      this_tab_done BOOLEAN := false;
      retry_count   INTEGER := 0;
      sleep_time    INTEGER := SLEEP_TIME_INITIAL;
      l_task_id     INTEGER;

      l_sql_del     VARCHAR2(2000);
      l_sql_ins     VARCHAR2(2000);
   BEGIN 
      
        l_sql_del := 'DELETE FROM rg_stored_dbh '
                        || 'WHERE summary_id = :1 '
                        || 'AND tstamp BETWEEN :2 AND ( :3 + 3599/86400) '
                        || 'AND instance_id IN (SELECT instance_id '
                        || 'FROM rg_busyhour_tmp '
                        || 'GROUP BY instance_id) ';

        l_sql_ins := 'INSERT INTO rg_stored_dbh '
                        || '(summary_id, instance_id, tstamp, rank, value) '
                        || 'SELECT :1, t.instance_id, t.tstamp, t.rank, t.value '
                        || 'FROM rg_busyhour_tmp t ';

      WHILE (NOT this_tab_done) LOOP
      
	 this_tab_done := true;
	 retry_count := retry_count + 1;
	 
	 BEGIN 
	    	    
            EXECUTE IMMEDIATE l_sql_del USING bh_def.id, startTime, endTime;
	    
            EXECUTE IMMEDIATE l_sql_ins USING bh_def.id;

	      INSERT INTO bh_refresh_sbh 
		(task_id, summary_id, instance_id, tstamp)
		SELECT task_id_in, bh_def.id, t.instance_id, t.tstamp
		  FROM rg_busyhour_tmp t
		  GROUP BY t.instance_id, t.tstamp;
				
	 EXCEPTION 
	    WHEN OTHERS THEN
	      this_tab_done := false;
	 END;
	 
	 this_tab_done := this_tab_done OR (retry_count > MAX_RETRY);
	 
	 IF (NOT this_tab_done) THEN
	    dbms_lock.sleep(sleep_time);
	    sleep_time := sleep_time * SLEEP_TIME_MULT;
	 END IF;
	    	 
      END LOOP;
      
      --
      -- If we exited because the retries were exhausted,
      -- add the table name to the list.
      -- 
      IF (retry_count > MAX_RETRY) THEN 
	 add_to_error_string('RG_STORED_DBH');
      END IF;
	 

   END UpdateRGStored;
   
   -- -----------------------------------------------------------------------
   -- PopulateCompleteTree - Expand the RG_BUSYHOUR_TMP table out to include
   -- all children and populate RG_BUSYHOUR_TREE_MV.
   -- -----------------------------------------------------------------------
   PROCEDURE PopulateCompleteTree(job_id_in INTEGER,
				  task_id_in INTEGER,
				  week_in DATE) IS
      cid INTEGER;
      i INTEGER;
      TYPE inst_tab_type IS TABLE OF rg_busyhour_tmp.instance_id%TYPE;
      inst_tab inst_tab_type;
      TYPE ent_tab_type IS TABLE OF bh_driver_tmp.entity_id%TYPE;
      ent_tab inst_tab_type;     
      start_tstamp DATE;  -- the earliest tstamp touched
      end_tstamp DATE;    -- the latest tstamp touched
      ex_end_tstamp DATE; -- the latest tstamp touched extented by 59m59s
      l_task_id INTEGER;
      l_has_frequency INTEGER;
   BEGIN

      --
      -- Cache the context_id for the reporting context.
      -- This isn't done for performance, but rather as a workaround for 
      -- Oracle's limitation that CONNECT BY queries can not contain a join.
      -- Starting with 9i this limitation is removed but introducing joins 
      -- for a single row lookup probably won't help much.
      --
      l_task_id := status.generic_start_task(job_id_in,'Starting initialization');
      
      SELECT context_id INTO cid FROM wmn_context WHERE context_use = 'R';
      
      SELECT min(tstamp), max(tstamp), max(tstamp)+g_fiftynines
	INTO start_tstamp, end_tstamp, ex_end_tstamp
	FROM bh_refresh_sbh r, active_task tsk
	WHERE tsk.job_id = job_id_in
	AND tsk.task_id = r.task_id
	AND r.tstamp >= week_in
	AND r.tstamp < week_in + 7;
      
      --
      -- Do BULK COLLECT to gather a unique list of instances that
      -- have been touched by this bhupdate job.  Since this collection
      -- can be pretty big, try to minimize the amount of time 
      -- we keep it around.  (ie: Don't intermingle other ops in
      -- between the BULK COLLECT and the DELETE.
      --
      SELECT r.instance_id BULK COLLECT INTO inst_tab
	FROM bh_refresh_sbh r, active_task tsk
	WHERE tsk.job_id = job_id_in
	AND tsk.task_id = r.task_id
	AND r.tstamp >= week_in
	AND r.tstamp < week_in + 7
	GROUP BY r.instance_id
	ORDER BY r.instance_id;
      
      status.generic_end_task(job_id_in, l_task_id,
	'Initialization complete.  Collected '||SQL%ROWCOUNT||' instances for '||
	to_char(start_tstamp,'YYYYMMDD HH24:MI:SS')||' to '||to_char(end_tstamp, 'YYYYMMDD HH24:MI:SS'));
      

      execute immediate 'TRUNCATE TABLE bh_driver_tmp';
      
      l_task_id := status.generic_start_task(job_id_in,'Starting populating instance tree');
      --
      -- Use the inst_tab array to populate RG_BUSYHOUR_TREE_TMP 
      -- with bhinstance_id (an instance_id that is the "root" for
      -- all instances under it)
      --
      FORALL i IN inst_tab.first..inst_tab.last
	INSERT INTO rg_busyhour_tree_tmp (instance_id, bhinstance_id, entity_id, parentinstance_id)
	SELECT instance_id, inst_tab(i) bhinst, entity_id, parentinstance_id
	  FROM wmn_instance_context
	  START WITH instance_id = inst_tab(i)
	  AND context_id = cid
	  CONNECT BY PRIOR instance_id = parentinstance_id
	  AND context_id = cid;

      status.generic_end_task(job_id_in, l_task_id, 'Populated '||to_number(SQL%ROWCOUNT)
        ||' instance(s) on the instance tree.');
	
	-- We don't need the table anymore...
	inst_tab.DELETE;
	
	
	SELECT entity_id BULK COLLECT INTO ent_tab
	  FROM rg_busyhour_tree_tmp
	  GROUP BY entity_id
	  ORDER BY entity_id;	
	
	FOR i IN ent_tab.first..ent_tab.last LOOP 
	  
	   l_task_id := status.generic_start_task(job_id_in,
	     'Starting driver population for entity '||ent_tab(i));
	   
	   SELECT count(*) INTO l_has_frequency
	     FROM rg_timeoffset toff, wmd_table_group wtg
	     WHERE toff.frequency_id = wtg.frequency_id
	     AND wtg.entity_id = ent_tab(i);
	   
	   IF (l_has_frequency > 0) THEN 
	      

	      INSERT INTO bh_driver_tmp (entity_id, tstamp, instance_id)
		SELECT x.ent, x.tstamp + off.offset, x.instance_id
		  FROM (SELECT /*+ ORDERED USE_HASH(tsk, rfrs) */
                        	ent_tab(i) ent, rfrs.tstamp, tree.instance_id
                         FROM active_task tsk,
                              bh_refresh_sbh rfrs,
                              rg_busyhour_tree_tmp tree
                         WHERE job_id_in = tsk.job_id
                           AND tsk.task_id = rfrs.task_id
                           AND tree.entity_id = ent_tab(i)
                           AND rfrs.instance_id = tree.bhinstance_id
                           AND rfrs.tstamp >= week_in
                           AND rfrs.tstamp < week_in + 7
                         GROUP BY rfrs.tstamp, tree.instance_id) x,
	              (SELECT /*+ NO_MERGE */ DISTINCT toff.seconds/(24*60*60) offset
                       FROM rg_timeoffset toff,
                             wmd_table_group wtg
                       WHERE toff.frequency_id = wtg.frequency_id
	                 AND wtg.entity_id = ent_tab(i)) off
	          GROUP BY x.ent, x.tstamp + off.offset, x.instance_id;
	   
          	status.generic_end_task(job_id_in, l_task_id,
	           'Populated '||SQL%ROWCOUNT||' rows for entity '||ent_tab(i));
	   ELSE 
	      
	      status.generic_end_task(job_id_in, l_task_id,
		'Short circuited entity '||ent_tab(i));
	      
	   END IF;

	END LOOP;
	   
	commit;
	
   END PopulateCompleteTree;
   
   -- -----------------------------------------------------------------------
   -- This generates the busyhour_task entries for the specific week
   -- Creates a busyhour_task for each table that requires a clean/copy
   -- -----------------------------------------------------------------------
   PROCEDURE SummarizeAllTables(
				job_id_in INTEGER,
				task_id_in INTEGER,
				week_in DATE) IS 

      l_start_date  DATE;
      l_end_date    DATE;

      p_task     BHtaskRec;
      bh_task_id INTEGER;

      start_id INTEGER;
      end_id   INTEGER;

   -- This is a tweaked version of BH_LIB but this one uses rcount
   -- largest first
   -- ... also count only distinct instance_ids as we slice on them
   CURSOR Entity_Tables(job_id_in INTEGER, i_period CHAR) IS
     SELECT /*+ NO_MERGE(bh) */ wdt.*, bh.rcount
       FROM ( SELECT entity_id, COUNT(distinct instance_id) rcount
              FROM bh_driver_tmp
              GROUP BY entity_id ) bh, 
       WMD_Dict_Table wdt
       WHERE wdt.is_time_series = 'Y'
       AND wdt.summary_id = 0
       AND (
       (summary_type.valid_DBH(wdt.summarize)=1 AND i_period='D')
       OR (summary_type.valid_WBH(wdt.summarize)=1 AND i_period='W')
       OR (summary_type.valid_MBH(wdt.summarize)=1 AND i_period='M')
       )
       AND wdt.entity_id = bh.entity_id
       ORDER BY bh.rcount DESC, wdt.entity_id;

   -- The NTILE MUST be used over a DISTINCT list or we run the
   -- risk of getting the same instance in 2 buckets
   CURSOR Entity_Slice(entity_in INTEGER, slice_in INTEGER) IS
      SELECT nt, MIN(instance_id) min_id, MAX(instance_id) max_id
      FROM
      ( SELECT instance_id, NTILE(slice_in) OVER (ORDER BY instance_id) nt
        FROM ( SELECT DISTINCT instance_id
               FROM bh_driver_tmp 
               WHERE entity_id = entity_in 
             )
      )
      GROUP BY nt;
	
   BEGIN
      
      
      --
      -- We know we have at least one time stamp for this week (see the
      -- valid_weeks cursor in GlobalUpdate), let's narrow it down to 
      -- the minimum span.  This is generally a Good Thing so we can
      -- do minimal spans of tables like BH_DRIVER_TMP and the TAB tables.
      -- When possible, avoid spilling over the weekly boundaries 
      -- since this would require an additional partition scan
      -- of the DBH tables (WEEKLY) or TAB tables (DAILY).
      -- This is only a concern of sub-hourly systems wher the 
      -- BH can be 23:15 and we can spill over to the day/week.
      -- 
      SELECT min(tstamp), max(tstamp)+g_fiftynines INTO l_start_date, l_end_date
	FROM bh_refresh_sbh r, active_task tsk
	WHERE tsk.job_id = job_id_in
	AND tsk.task_id = r.task_id
	AND r.tstamp >= week_in
	AND r.tstamp < week_in + 7;

      -- This is a tweaked cursor that will generate the entity list
      -- by the count of instance_ids per entity, descending. This gives
      -- us the biggest group first.
      FOR table_def IN Entity_Tables(job_id_in, 'D') LOOP

         -- Load the record
         p_task.job_id := job_id_in;             -- job id
         p_task.status_code := g_OPS_READY;      -- READY
         p_task.start_date := l_start_date;
         p_task.end_date := l_end_date;
         p_task.retry_count := 0;                -- number of retry attempts
         p_task.table_id := table_def.table_id;
         p_task.table_name := table_def.table_name;
         p_task.entity_id := table_def.entity_id;
         p_task.is_time_series := table_def.is_time_series;
         p_task.summary_id := table_def.summary_id;
         p_task.start_id := start_id;
         p_task.end_id := end_id;
         
         -- Only slice when needed ... 
         -- and make sure that all slices have at least 1 instance_id
         IF ( table_def.rcount >= g_slicethreshold  
              AND (table_def.rcount / g_slices > 1.0)
              AND (g_slicethreshold > 0 AND g_slices > 1) ) THEN

            -- Slices are needed
            FOR slice_def IN Entity_Slice(table_def.entity_id, g_slices) LOOP

               -- We need a new task for each table
               bh_task_id := Status.generic_start_task( job_id_in, 'ADD: ' || p_task.table_name );
 
               p_task.task_id := bh_task_id;           -- task id for this specific task

               -- Add the start and end
               p_task.start_id := slice_def.min_id;
               p_task.end_id := slice_def.max_id;

               -- Add the busyhour_task
               add_bh_task( job_id_in, bh_task_id, p_task );

            END LOOP;

         ELSE

         -- We need a new task for each table
         bh_task_id := Status.generic_start_task( job_id_in, 'ADD: ' || p_task.table_name );
 
         p_task.task_id := bh_task_id;           -- task id for this specific task

         -- Add the busyhour_task
         add_bh_task( job_id_in, bh_task_id, p_task );
         END IF;

      END LOOP;

   END SummarizeAllTables;
   
   
   -- -----------------------------------------------------------------------
   -- SummarizeTable - Do the BH store for one table
   -- -----------------------------------------------------------------------
   FUNCTION SummarizeTable( table_desc BHtaskRec ) RETURN BOOLEAN 
   IS
      selStr VARCHAR2(16000);
      insStr VARCHAR2(16000);
      lv_merge_str varchar2(32000);
      lv_merge_use_str varchar2(4000);
      lv_update_list VARCHAR2(16000);
      lv_insert_param_list VARCHAR2(16000);
      lv_insert_value_list VARCHAR2(16000);

      timerStart NUMBER;
      timerEnd   NUMBER;
      sDateTime  DATE;
      eDateTime  DATE;
      lv_start_id INTEGER := 0;
      lv_end_id INTEGER := 99999999999999;
      lv_updcount INTEGER := 0;
      retVal     BOOLEAN := true;

      l_task_id INTEGER;
      l_err_str VARCHAR2(600);
      l_bh_table VARCHAR2(30);

      -- Cache the dates for reuse
      c_start_date VARCHAR2(30);
      c_end_date VARCHAR2(30);

   BEGIN
      
      -- Cache the dates for reuse
      c_start_date := to_char(table_desc.start_date, 'YYYYMMDD HH24:MI:SS');
      c_end_date := to_char(table_desc.end_date, 'YYYYMMDD HH24:MI:SS');
      
      l_bh_table := BH_Lib.GetBHDestName('D', table_desc.table_name);

      -- Are we using these?
      sDateTime := sysdate;
      timerStart := dbms_utility.get_time;
                  
      IF ( table_desc.start_id IS NOT NULL ) THEN
        lv_start_id := table_desc.start_id;
        lv_end_id := table_desc.end_id;
      END IF;

      insStr := 'INSERT INTO '||l_bh_table||
	'(summary_id, instance_id, tstamp';
           
      selStr := 'SELECT /*+ USE_HASH(d t) */ -1, t.instance_id, t.tstamp';

      lv_insert_param_list := 'i.summary_id, i.instance_id, i.tstamp';
      lv_insert_value_list := '-1,ms.instance_id,ms.tstamp';
      
      FOR field_def IN BH_Lib.Table_Fields(table_desc.table_id) LOOP
	 selStr := selStr || ',' || field_def.table_column;
	 insStr := insStr || ',' || field_def.table_column;	
         if ( lv_updcount = 1 ) then
           lv_update_list := lv_update_list||',';
         else
           lv_updcount := 1;
         end if;
         lv_update_list := lv_update_list||'i.'||field_def.table_column||'=ms.'||field_def.table_column;
         lv_insert_param_list := lv_insert_param_list||',i.'||field_def.table_column;
         lv_insert_value_list := lv_insert_value_list||',ms.'||field_def.table_column;
      END LOOP;
      
      insStr := insStr || ') ';
      selStr := selStr || ' FROM bh_driver_tmp d, '||
        table_desc.table_name||
	' t' ||
	' WHERE t.instance_id = d.instance_id' ||
	' AND t.tstamp = d.tstamp' ||
	' AND d.entity_id = :entity_id' ||
	' AND d.tstamp BETWEEN :start_date AND :end_date' ||
	' AND t.tstamp BETWEEN :start_date AND :end_date';

      selStr := selStr || ' AND d.instance_id BETWEEN :start_id AND :end_id';

      lv_merge_str := 'MERGE /*+ USE_HASH(ms i) */ INTO '||l_bh_table||' i '||
                      'USING ('||selStr||' ) ms '||
                      'ON (i.summary_id = -1 '||
                      'AND i.tstamp = ms.tstamp '||
                      'AND i.instance_id = ms.instance_id) '||
                      'WHEN MATCHED THEN UPDATE SET '||lv_update_list||
                      ' WHEN NOT MATCHED THEN INSERT ('||lv_insert_param_list||
                      ') VALUES ('||lv_insert_value_list||')';
      
      lv_merge_use_str := ' table_desc.entity_id,
                          table_desc.start_date,
                          table_desc.end_date,
                          table_desc.start_date,
                          table_desc.end_date,
			  lv_start_id,
			  lv_end_id';

      BEGIN

         dbms_application_info.set_client_info( 'Clean: ' 
            || l_bh_table 
            || slice_msg( table_desc.start_id, table_desc.end_id )
            );

         -- End the original task to show that we have ACQUIRED
         status.generic_end_task(table_desc.job_id, table_desc.task_id,
            'ACQUIRED: ' 
            || table_desc.table_name
            || slice_msg( table_desc.start_id, table_desc.end_id )
            );

/*
	 CleanOldBH(table_desc.job_id, 
	   table_desc.task_id,
	   table_desc.table_name,
	   table_desc.entity_id,
	   table_desc.start_date,
	   table_desc.end_date,
           table_desc.start_id,
           table_desc.end_id);
*/
	 
         -- We are marking each COPY with dates which is great if we have more than 1 week
	 l_task_id := status.generic_start_task(table_desc.job_id, 'Copying '
                         || table_desc.table_name
                         || ' between '
                         || c_start_date
                         || ' and '
                         || c_end_date
                         || slice_msg( table_desc.start_id, table_desc.end_id )
                         );
	 
         dbms_application_info.set_client_info( 'Copy: ' 
            || l_bh_table
            || slice_msg( table_desc.start_id, table_desc.end_id )
            );

       BEGIN

         if ( gOverRide = 'Y' ) then
            execute immediate lv_merge_str USING
	      table_desc.entity_id,
	      table_desc.start_date,
	      table_desc.end_date,
	      table_desc.start_date,
	      table_desc.end_date,
	      lv_start_id,
	      lv_end_id;
         else
	   execute immediate insStr || selStr using
	     table_desc.entity_id,
	     table_desc.start_date,
	     table_desc.end_date,
	     table_desc.start_date,
	     table_desc.end_date,
	     lv_start_id,
	     lv_end_id;
         END IF;

        EXCEPTION
          WHEN DUP_VAL_ON_INDEX then
            execute immediate lv_merge_str USING
	      table_desc.entity_id,
	      table_desc.start_date,
	      table_desc.end_date,
	      table_desc.start_date,
	      table_desc.end_date,
	      lv_start_id,
	      lv_end_id;
          WHEN OTHERS THEN
	    l_err_str := 'FAILED '||l_bh_table||' '||SQLERRM;
	    rollback;	   
	    l_task_id := status.generic_start_task(table_desc.job_id, 'Copying '||l_bh_table);
	    status.generic_end_task(table_desc.job_id, table_desc.task_id, l_err_str, FALSE);
	    commit;
	    retVal := false;
       END; 
         -- We are marking each COPY with dates which is great if we have more than 1 week
	 status.generic_end_task(table_desc.job_id, l_task_id, 'Copied '
            || l_bh_table
            || ':'
            || SQL%ROWCOUNT
            || ' rows between '
            || c_start_date
            || ' and '
            || c_end_date
            || slice_msg( table_desc.start_id, table_desc.end_id )
            );

	 COMMIT;
	   
      EXCEPTION
	 WHEN OTHERS THEN
	   l_err_str := 'FAILED '||l_bh_table||' '||SQLERRM;
	   rollback;	   
	   l_task_id := status.generic_start_task(table_desc.job_id, 'Copying '||l_bh_table);
	   status.generic_end_task(table_desc.job_id, table_desc.task_id, l_err_str, FALSE);
	   commit;
	   retVal := false;
      END;	

      -- Are we using these?
      timerEnd := dbms_utility.get_time;
      eDateTime := sysdate;
      
      RETURN retVal;
      
   END SummarizeTable;
			   
   
   -- -----------------------------------------------------------------------
   -- CleanOldBH - Delete any old BHs that are already in the table that
   -- are about to be overlaid.
   -- -----------------------------------------------------------------------
   PROCEDURE CleanOldBH(job_id_in      IN INTEGER,
			task_id_in     IN INTEGER,			
			clean_table    IN VARCHAR2,
			table_entity   IN INTEGER,
			startTime      IN DATE,
			endTime        IN DATE,
                        start_id       IN INTEGER,
                        end_id         IN INTEGER) IS
      sqlStr   VARCHAR2(4000);
      bh_table VARCHAR2(30);
      l_task_id INTEGER;

      -- Are we using?
      timerStart NUMBER;
      timerEnd   NUMBER;

   BEGIN
      
      -- Are we using?
      timerStart := dbms_utility.get_time;

      bh_table := BH_Lib.GetBHDestName('D', clean_table);
      l_task_id := status.generic_start_task(job_id_in, 'Cleaning '
         || bh_table
         || slice_msg( start_id, end_id )
         );
      

      sqlStr := 'DELETE FROM '||
	' (SELECT /*+ USE_HASH(dbh d) */ dbh.instance_id, dbh.instance_id, dbh.tstamp '||
	' FROM '||bh_table||' dbh, bh_driver_tmp d '||
	' WHERE dbh.summary_id = -1 '||
	' AND dbh.tstamp BETWEEN :sTime AND :eTime '||
	' AND dbh.instance_id = d.instance_id '||
	' AND dbh.tstamp = d.tstamp '||
	' AND d.entity_id = :entity_id_in ';

      IF ( start_id IS NOT NULL ) THEN
        sqlStr := sqlStr || 
          ' AND d.instance_id BETWEEN :start_id AND :end_id'||
	  ' AND d.tstamp BETWEEN :sTime AND :eTime)';
      ELSE
        sqlStr := sqlStr || 
	  ' AND d.tstamp BETWEEN :sTime AND :eTime)';
      END IF;
      
      BEGIN 

         IF ( start_id IS NOT NULL ) THEN
            execute immediate sqlStr using
               startTime, endTime, table_entity, start_id, end_id, startTime, endTime;
         ELSE
            execute immediate sqlStr using
               startTime, endTime, table_entity, startTime, endTime;
         END IF;
	 
	 status.generic_end_task(job_id_in, l_task_id, 'Cleaned '
            || SQL%ROWCOUNT
            || ' rows from '
            || bh_table
            || slice_msg( start_id, end_id )
            );

      EXCEPTION 
	 WHEN OTHERS THEN
	   status.generic_end_task(job_id_in, l_task_id,
	   'FAILED: Clean from '||bh_table||' '||SQLERRM, FALSE);	   
      END; 
	 
      -- Are we using?
      timerEnd := dbms_utility.get_time;
      
   END CleanOldBH;
   
   -- -----------------------------------------------------------------------
   -- CreateBHTables - Create a SBH version of the table.  The table is a
   -- weekly partitioned IOT with field types changed based on the aggr type.
   -- Only one partition is created based on sysdate.  Partition maintenance
   -- can be used to push out these partitions to whatever is needed for
   -- production use.
   -- -----------------------------------------------------------------------
--   This procedure is moved to CreateBHTab package in file CreateBHTab.sql
   -- -----------------------------------------------------------------------
   --
   -- -----------------------------------------------------------------------

   PROCEDURE TruncateBHTables(period CHAR) IS
   BEGIN
      MaintainBHTables(period, 'TRUNCATE');
   END TruncateBHTables;
   
   -- -----------------------------------------------------------------------
   --
   -- -----------------------------------------------------------------------
   PROCEDURE DropBHTables(period CHAR) IS
   BEGIN
      MaintainBHTables(period, 'DROP');
   END DropBHTables;
   
   -- -----------------------------------------------------------------------
   --
   -- -----------------------------------------------------------------------
   PROCEDURE AnalyzeBHTables(period CHAR) IS
   BEGIN
      MaintainBHTables(period, 'ANALYZE');
   END AnalyzeBHTables;
   
   -- -----------------------------------------------------------------------
   -- MaintainBHTables - Maintain all of the tables for a particular period.
   -- Tables can be truncated, dropped or analyzed.
   -- -----------------------------------------------------------------------
   PROCEDURE MaintainBHTables(period CHAR, op VARCHAR2 DEFAULT NULL) IS
      sqlStr    VARCHAR2(4000);
      tableName VARCHAR2(40);
   BEGIN
      

      FOR table_def IN BH_Lib.Raw_Entity_Tables(period) LOOP
	 
	 tableName := BH_Lib.GetBHDestName(period, table_def.table_name);
	 
	 IF (op = 'TRUNCATE') THEN
	    sqlStr := 'TRUNCATE TABLE '||tableName;
	 ELSIF (op = 'DROP') THEN
	    sqlStr := 'DROP TABLE '||tableName;
	 ELSIF (op = 'ANALYZE') THEN 
	    sqlStr := 'ANALYZE TABLE '||tableName||' COMPUTE STATISTICS';
	 ELSE 
	    sqlStr := NULL;
	 END IF;
	 
	 BEGIN
	    IF (sqlStr IS NOT NULL) THEN 
	       execute immediate sqlStr;
	    ELSE
	       dbms_output.put_line(table_def.table_name);
	    END IF;
	 EXCEPTION
	    WHEN OTHERS THEN
	      dbms_output.put_line('ERROR: ' ||
	      BH_Lib.GetBHDestName(period, table_def.table_name));
	      
	      dbms_output.put_line(' '||SQLERRM);
	 END;
	 
	 
      END LOOP;
      
   END MaintainBHTables;
        
   -- -----------------------------------------------------------------------
   -- AlterTunables
   -- -----------------------------------------------------------------------
   PROCEDURE AlterTunables IS
      v INTEGER;
   BEGIN
      
      init_tunable_parameters;

      -- SEAde50080 for 9i only, 8i will ignore this
      wm_session.workarea_size_policy_manual;

      execute immediate 'ALTER SESSION SET sort_area_size = '
	||to_char(g_sort_area_size);
   
      execute immediate 'ALTER SESSION SET sort_area_retained_size = '
	||to_char(g_sort_area_retained_size);

      execute immediate 'ALTER SESSION SET hash_area_size = '
	||to_char(g_hash_area_size);
   
   EXCEPTION
      WHEN OTHERS THEN
	NULL;   -- Oh well, at least we tried...	
   END AlterTunables;
   
   -- -----------------------------------------------------------------------
   --
   -- -----------------------------------------------------------------------
   PROCEDURE RGLocking(task_id_in INTEGER, start_or_stop VARCHAR2) IS
   BEGIN
      
      IF (upper(start_or_stop) = 'START') THEN 
	 INSERT INTO rg_data_depend (taskid, instance_id, firsttime, lasttime, loadersource_id)
	   SELECT taskid, instance_id, firsttime, lasttime, loadersource_id + 20
	     FROM rg_data_depend
	     WHERE taskid = task_id_in;
      ELSE
	 DELETE FROM rg_data_depend
	   WHERE taskid = task_id_in AND loadersource_id >= 20;
      END IF;
      
      commit;
      
   END RGLocking;
   
   
   -- -----------------------------------------------------------------------
   -- CleanOldPartitions - Move FirstAvailable when the partitions get
   -- purged.  If you change this chances are you need to change 
   -- summary.CleanOldPartitions
   -- -----------------------------------------------------------------------
   PROCEDURE CleanOldPartitions(table_name VARCHAR2, tstamp DATE) IS
      new_first DATE;
      new_last  DATE;
      new_last_pabh DATE;
      new_first_pabh DATE;
      new_first_needed DATE;

      vSpan varchar2(5);
      
      cursor all_bh(pSpan varchar2) IS 
	SELECT pmb.firstneeded pmb_fn, pmb.firstavail pmb_fa,pmb.lastavail pmb_la,pmb.id pmb_id,
	  pab.firstavail pab_fa,pab.lastavail pab_la
	  FROM PM_BusyHour pmb, pm_pabh pab
	  WHERE pmb.field_id IS NOT NULL
	  AND pmb.span=pSpan
	  AND pmb.id=pab.id(+)
	  FOR UPDATE;
   BEGIN
      
      if table_name like '%DBH' then 
         vSpan := 'D';
      elsif table_name like '%WBH'
            or table_name= 'RG_STORED_PABH' then
         vSpan := 'W';
      elsif table_name like '%WBH' then
	 vSpan := 'M';
      else
	raise_application_error(-20100, 'Table '||table_name
	   ||' has wrong suffix!!!, check part_tables.');
      end if;

      FOR this_bh IN all_bh(vSpan) LOOP
	 
	 if vSpan = 'D' or vSpan = 'W' then
	   new_first := tstamp + 7;
	 else
           new_first := add_months(tstamp,1);
         end if;
	 new_last := this_bh.pmb_la;
         new_first_needed := new_first;

	 new_first_pabh := new_first;
	 new_last_pabh := this_bh.pab_la;

	 -- Check for firstavail past lastavail in pm_busyhour
	 IF (nvl(new_first, to_date('1/1/2000', 'MM/DD/YYYY')) > 
	   nvl(new_last, to_date('1/1/1999', 'MM/DD/YYYY'))) THEN
	    new_first := NULL;
	    new_last := NULL;
	 END IF;
	 
	 -- Check for firstavail past lastavail in pm_pabh
	 IF (nvl(new_first, to_date('1/1/2000', 'MM/DD/YYYY')) >
	   nvl(new_last_pabh, to_date('1/1/1999', 'MM/DD/YYYY'))) THEN
	    new_first_pabh := NULL;
	    new_last_pabh := NULL;
	 END IF;
	 
	UPDATE PM_BusyHour
	   SET 
	   firstavail = greatest(new_first,
               nvl(firstavail, to_date('1/1/1970', 'MM/DD/YYYY'))),
	   firstneeded = nvl(
               greatest(new_first_needed,
               nvl(firstneeded, to_date('1/1/1970', 'MM/DD/YYYY'))),
                         firstneeded),
	   lastavail = new_last
	   WHERE id = this_bh.pmb_id;
	 
	 update pm_pabh
	   set 
	   firstavail = greatest(new_first_pabh,
               nvl(firstavail,trunc(to_date('1/1/1970','MM/DD/YYYY'),'D'))),
	   lastavail = new_last_pabh
	   where id = this_bh.pmb_id;
	 
      END LOOP;
      
   END CleanOldPartitions;
   
   
   PROCEDURE update_task(task_id_in INTEGER, message VARCHAR2) IS
      job_id_in INTEGER;
   BEGIN
      
      SELECT job_id INTO job_id_in
	FROM active_task
	WHERE task_id = task_id_in;
      
      dbms_output.put_line(job_id_in || '/' || task_id_in || ' ' || message);
      status.update_status(job_id_in, task_id_in, sched_general.TM_Accepted, message);
      
   END update_task;
   
   -----------------------------------------------------------------
   -- This proc. checks for the validity of various arguments passed
   -----------------------------------------------------------------   
   PROCEDURE validate_and_insert_bh_params(i_forward_flag CHAR,
					   i_backward_flag CHAR,
					   i_skip_flag CHAR,
					   i_start_date IN VARCHAR2,
					   i_end_date IN  VARCHAR2,
					   i_entity_name IN  VARCHAR2,
					   i_override IN CHAR,
					   o_feedback_id IN OUT NUMBER,
					   i_is_busyhour IN CHAR,
					   pBH_name varchar2,
					   pBH_id integer,
                                           pSpan  varchar2 DEFAULT NULL) IS
      i_status_code task_status.status_code%TYPE:=0;
      i_ended task_status.ended%TYPE:=NULL;

      invalid_override EXCEPTION;
      invalid_start_date EXCEPTION;
      invalid_end_date EXCEPTION;
      invalid_date_combo EXCEPTION;
      missing_date_combo EXCEPTION;
      invalid_flag_combo EXCEPTION;
      invalid_entity EXCEPTION;
      invalid_forward_date_combo EXCEPTION;
      invalid_backward_date_combo EXCEPTION;
      entity_does_not_exist EXCEPTION;
      invalid_bh_id EXCEPTION;
      invalid_span EXCEPTION;
      bhdef_does_not_exist EXCEPTION;
      i_entity_id INTEGER:=null;
      i_entity_id_bh INTEGER;
      i_start_date_local DATE;
      i_end_date_local DATE;
      i_start_date1 VARCHAR2(50);
      i_end_date1 VARCHAR2(50);
      i_entity_name1 VARCHAR2(30);
      i_sched_id INTEGER;
      i_bh_id INTEGER;
      i_job_id NUMBER;
      new_job_type INTEGER;
      new_sched_fk INTEGER;
      iter_count   INTEGER;
      new_feedback_id INTEGER;
      vBH_id       integer;
      vErr_msg     varchar2(128) := null;
   BEGIN 
      -- Check for the validity of arguments
      IF i_forward_flag = 'Y' AND i_backward_flag = 'Y' THEN
	 RAISE invalid_flag_combo;
      END IF;
      
      IF i_start_date='N' THEN
	 i_start_date1:=NULL;
      ELSE
	 i_start_date1:=i_start_date;
      END IF;
      IF i_end_date='N' THEN
	 i_end_date1:=NULL;
      ELSE
	 i_end_date1:= i_end_date;	
      END IF;
      
      IF (i_start_date1 IS NULL AND i_end_date1 IS NOT NULL) OR
	(i_start_date1 IS NOT NULL AND i_end_date1 IS NULL)
      THEN
	 RAISE missing_date_combo;
      END IF;

   -- override requires both -d and -D
   IF (i_override = 'Y' AND (i_start_date1 IS NULL OR i_end_date1 IS NULL)) THEN
      RAISE invalid_override;
   END IF;

      -- Check for the validity of start_date
      IF (i_start_date1 IS NOT NULL AND i_end_date1 IS NOT NULL) THEN
   BEGIN
	    SELECT to_date(i_start_date1,'yyyymmdd')
	      INTO i_start_date_local
	      FROM dual;
   EXCEPTION
      WHEN OTHERS THEN
	RAISE invalid_start_date;
   END;
   -- Check for the validity of end_date
   BEGIN
      SELECT to_date(i_end_date1,'yyyymmdd')
	INTO i_end_date_local
	FROM dual;
   EXCEPTION
      WHEN OTHERS THEN
	RAISE invalid_end_date;
   END;
   
   -- Check for the validity of date combo
   IF to_date(i_start_date1,'yyyymmdd') > to_date(i_end_date1,'yyyymmdd') THEN
      RAISE invalid_date_combo;
   END IF;
   -- Check for the invalid combo of -f and date flags
   IF i_forward_flag='Y' AND  i_start_date1 IS NOT NULL AND i_end_date1 IS NOT NULL THEN
      RAISE invalid_forward_date_combo;
   END IF;
   -- Check for the invalid combo of -b and date flags
   IF i_backward_flag='Y' AND  i_start_date1 IS NOT NULL AND i_end_date1 IS NOT NULL THEN
      RAISE invalid_backward_date_combo;
   END IF;

   
      END IF;

   -- check for busyhour id
   if pBH_Id is not null then
     begin
       select id into vBH_id
         from pm_busyhour
        where id = pBH_Id
	AND (   ((i_is_busyhour = 'Y') AND (field_id IS NOT NULL))
	     OR ((i_is_busyhour = 'N') AND (field_id IS NULL)) );
     exception
       when no_data_found then
        raise invalid_bh_id;
     end;
   else
     -- check for table type
     if pSpan not in ('D','W', 'M') and pSpan is not null then
       raise invalid_span;
     end if;

      -- Check for the validity of entity_name
      IF i_entity_name='N' THEN
	 i_entity_name1:=NULL;
      ELSE
	 i_entity_name1:=i_entity_name;
      END IF;

      IF i_entity_name1 IS NOT NULL then
        BEGIN
          SELECT entity_id
            INTO i_entity_id
            FROM wmn_entity
            WHERE upper(entity_name)=upper(i_entity_name1);
        EXCEPTION
            WHEN no_data_found THEN
              RAISE invalid_entity;
        END;
      ELSE
        i_entity_id := null;
      END IF;
   
        -- Check for the existence of entity_name in pm_busyhour table
        -- with the -e, -s and -n options
        BEGIN
          SELECT cluster_entity_id
            INTO i_entity_id_bh
            FROM pm_busyhour
            WHERE cluster_entity_id=nvl(i_entity_id, cluster_entity_id)
              and name = nvl(pBH_name, name)
              and span = nvl(pSpan, span)
            AND (   ((i_is_busyhour = 'Y') AND (field_id IS NOT NULL))
                 OR ((i_is_busyhour = 'N') AND (field_id IS NULL)) )
            AND ROWNUM=1;
        EXCEPTION
          WHEN no_data_found THEN
            RAISE BHdef_does_not_exist;
        END;
      
   end if;

      SELECT bh_parameter_seq.nextval
	INTO i_bh_id
	FROM dual;
      
      INSERT INTO bh_parameter
	(id,
	entity_id,
	forward,
	backward,
	start_date,
	end_date,
	skip_missing,
	override,
	bh_name,
	bh_id,
        span)
	VALUES
	(i_bh_id,
	i_entity_id,
	i_forward_flag,
	i_backward_flag,
	i_start_date_local,
	i_end_date_local,
	i_skip_flag,
	i_override,
	pBH_name,
	pBH_id,
        pSpan);
      
      SELECT schedule_seq.nextval
	INTO i_sched_id
	FROM dual;
      
      IF (i_is_busyhour = 'Y') THEN
	 new_job_type := 30;
	 new_sched_fk := i_bh_id;
      ELSE 
	 INSERT INTO TM_JobDesc
	   (jobdesc_id, software_id, parameter_fk)
	   VALUES
	   (jobdesc_seq.nextval, 3, i_bh_id)
	   returning jobdesc_id INTO new_sched_fk;
	 new_job_type := 25;
      END IF;
      
      INSERT INTO schedule
	(schedule_id,
	job_type,
	start_time,
	end_time,
	period,
	frequency,
	sched_fk) values
	(i_sched_id,
	new_job_type,
	SYSDATE,
	null,
	'I',
	1,
	new_sched_fk);
      commit;
      
      -- Get the job_id
      -- We need to poll here because of latency in the scheduler
      i_job_id := -1;
      iter_count := 0;    
      WHILE (i_job_id < 0 AND iter_count < 6) LOOP
	 iter_count := iter_count + 1;
	 BEGIN    
	    SELECT job_id
	      INTO i_job_id
	      FROM active_job
	      WHERE sched_fk= new_sched_fk
	      AND job_type = new_job_type;
	 EXCEPTION 
	    WHEN OTHERS THEN 
	      i_job_id := -1;
	      dbms_lock.sleep(5);
	 END;
      END LOOP;
      
      IF (i_job_id < 0) THEN 
	 raise_application_error(-20002, 'Can''t find scheduled job');
      END IF;
      
      
      -- insert job_id into bhupdate_feedback table
      -- but make sure the old one is gone
      SELECT bh_feedback_seq.nextval INTO new_feedback_id
	FROM dual;
      
      BEGIN 
	 DELETE FROM bh_feedback
	   WHERE id = new_feedback_id;
      EXCEPTION
	 WHEN OTHERS THEN
	   NULL; -- no big deal, it's supposed to happen
      END;
      
      INSERT INTO bh_feedback
	(id,job_id)
	VALUES
	(new_feedback_id,i_job_id)
	returning id INTO o_feedback_id;
      commit;
      
      -- dbms_output.put_line('fb id '|| i_job_id);
      
   EXCEPTION
      WHEN invalid_flag_combo THEN
	dbms_output.put_line('======================================================= ');
	dbms_output.put_line('===== Error : Cannot specify both -f and -b flags ===== ');
	dbms_output.put_line('======================================================= ');
	o_feedback_id := 0;
	
      WHEN missing_date_combo THEN
	dbms_output.put_line('================================================================ ');
	dbms_output.put_line('===== Error : Must specify both -d and -D flags or neither ===== ');
	dbms_output.put_line('================================================================ ');   
	o_feedback_id := 0;
	
      WHEN invalid_start_date THEN
	dbms_output.put_line('===================================================================== ');
	dbms_output.put_line('===== Error : Must specify start date in the format of YYYYMMDD ===== ');
	dbms_output.put_line('===================================================================== ');
	o_feedback_id := 0;
	
      WHEN invalid_end_date THEN
	dbms_output.put_line('=================================================================== ');
	dbms_output.put_line('===== Error : Must specify end date in the format of YYYYMMDD ===== ');
	dbms_output.put_line('=================================================================== ');
	o_feedback_id := 0;
	
      WHEN invalid_date_combo THEN
	dbms_output.put_line('======================================================================= ');
	dbms_output.put_line('===== Error : Start_date should be equal to or less than end_date ===== ');
	dbms_output.put_line('======================================================================= ');
	o_feedback_id := 0;
	
      WHEN invalid_entity THEN
	dbms_output.put_line('================================================ ');
	dbms_output.put_line('===== Error : Specify a valid entity name. ===== ');
	dbms_output.put_line('================================================ ');
	o_feedback_id := 0;
	
      WHEN entity_does_not_exist  THEN  
	IF (i_is_busyhour = 'Y') THEN 
	   dbms_output.put_line('======================================================================= ');
	   dbms_output.put_line('===== Error : Given entity does not have stored busy hour defined ===== ');
	   dbms_output.put_line('======================================================================= ');
	ELSE 
	   dbms_output.put_line('===================================================================== ');
	   dbms_output.put_line('===== Error : Given entity does not have stored summary defined ===== ');
	   dbms_output.put_line('===================================================================== ');
	END IF;
	o_feedback_id := 0;
	
      WHEN invalid_forward_date_combo THEN
	dbms_output.put_line('=============================================================== ');
	dbms_output.put_line('===== Error : Cannot specify both -f flag and -d/-D flags ===== ');
	dbms_output.put_line('=============================================================== ');
	o_feedback_id := 0;
	
      WHEN invalid_backward_date_combo THEN
	dbms_output.put_line('=============================================================== ');
	dbms_output.put_line('===== Error : Cannot specify both -b flag and -d/-D flags ===== ');	
	dbms_output.put_line('=============================================================== ');
	o_feedback_id := 0;

      WHEN invalid_override THEN
	dbms_output.put_line('============================================================= ');
	dbms_output.put_line('===== Error : Cannot specify -o without the -d/-D flags ===== ');	
	dbms_output.put_line('============================================================= ');
	o_feedback_id := 0;
	
      WHEN invalid_bh_id then
        dbms_output.put_line('======================================================================= ');
        dbms_output.put_line('===== Error : BH/SM ID ('||to_char(pBH_id)||') doesn''t exist ===== ');
        dbms_output.put_line('======================================================================= ');
	o_feedback_id := 0;

      WHEN invalid_span then
        dbms_output.put_line('============================================ ');
        dbms_output.put_line('===== Error : Invalid Span '||pSpan||' ===== ');
        dbms_output.put_line('============================================ ');
	o_feedback_id := 0;

      WHEN bhdef_does_not_exist then
        if i_entity_name1 is not null 
         or pSpan is not null 
         or pBH_name is not null then
           verr_msg := 'with ';
           if i_entity_name1 is not null then
              verr_msg := verr_msg||' entity_name='||i_entity_name||' ';
           end if;
           if pBH_name is not null then
              verr_msg := verr_msg||' BH_name='||pBH_name||' ';
           end if;
           if pSpan is not null then
              verr_msg := verr_msg||' Span='||pSpan||' ';
           end if;
        end if;
        dbms_output.put_line('============================================================ ');
        dbms_output.put_line('===== Error : No BH/SM definition '||verr_msg||' ===== ');
        dbms_output.put_line('============================================================ ');
	o_feedback_id := 0;

   END validate_and_insert_bh_params;

   -----------------------------------------------------------------
   -- ins_and_del - Cleans up bh_feedback for the current SBH job
   -- and schedules a kill job for bhupdate
   -----------------------------------------------------------------
   PROCEDURE ins_and_del_sm(fid NUMBER) IS
   BEGIN
      ins_and_del(fid, 'N');
   END ins_and_del_sm;
   
   
   PROCEDURE ins_and_del(fid NUMBER, i_is_busyhour CHAR DEFAULT 'Y') IS
   BEGIN
      
      -- schedule a kill job for bhupdate/smupdate
      insert into schedule
	(schedule_id,job_type,start_time,frequency,sched_fk,period,admin_task)
	select 
	schedule_seq.nextval,decode(i_is_busyhour, 'Y', 30, 20) ,sysdate,1,bhf.job_id,'I',2
	  from bh_feedback bhf
	  where bhf.id=fid;
	-- Cleanup bh_feedback
	delete from bh_feedback
	  where id=fid;
	commit;
   END ins_and_del;
   
   -----------------------------------------------------------------
   -- poll - Polls task status table to output status of bhupdate
   -- periodically. This is called from bhupdate.sh
   -----------------------------------------------------------------
   PROCEDURE poll(i_feedback_id IN NUMBER, o_status_id IN OUT NUMBER)
   IS
      i_task_id NUMBER := 0;
      i_job_id NUMBER := 0;
      i_status_code NUMBER := 0;
      i_sc     NUMBER := 0;
      i_max_task_id NUMBER := 0;
      prev_task_id NUMBER := 0;
      prev_msg VARCHAR2(600) := NULL;
      i_ts_msg VARCHAR2(600) := NULL;
      min_date DATE := NULL;
      max_date DATE := NULL;
      task_in_process NUMBER := 0;
      sucess_run EXCEPTION;
      failed_run EXCEPTION;
      scheduled EXCEPTION;
      waiting EXCEPTION;
      pending EXCEPTION;
      deleted EXCEPTION;
      
   BEGIN
      -- get the job_id for the current feedback id
      BEGIN
	 SELECT job_id
	   INTO i_job_id
	   FROM bh_feedback
	   WHERE id=i_feedback_id;
      EXCEPTION
	 WHEN no_data_found THEN
	   dbms_output.put_line('No data found for the job_id in bh_feedback.');
           o_status_id := 255; 
           return;
	   -- Get the task_id for current processing job
      END;
      BEGIN
	 SELECT nvl(max(task_id),0)
	   INTO i_task_id
	   FROM task_status
	   WHERE job_id=i_job_id;
      EXCEPTION
	 WHEN no_data_found THEN
	   null;
	   --dbms_output.put_line('NDF for job_id in ts '|| i_job_id);
      END;
      
      IF i_task_id=0 THEN
	 BEGIN
	    -- Get the status of the job
	    SELECT status_code
	      INTO i_status_code
	      FROM task_status
	      WHERE job_id=i_job_id;
	 EXCEPTION
	    WHEN no_data_found THEN
	      NULL;
	      --dbms_output.put_line('NDF for job_id for max sc ' ||i_job_id);
	 END;
      ELSE
	 BEGIN
	    -- Get the status of the job
	    SELECT status_code
	      INTO i_status_code
	      FROM task_status
	      WHERE job_id=i_job_id
	      AND task_id=i_task_id;
	 EXCEPTION
	    WHEN no_data_found THEN
	      NULL;
	      --dbms_output.put_line('NDF for job_id for max sc ' ||i_job_id);
	 END;
      END IF;
      
      IF i_status_code=6 THEN
	 RAISE deleted;
      END IF;
      -- Get the previous task id and message from the bhupdate_feedback table
      
      BEGIN
	 SELECT task_id,nvl(substr(message,1,500),'blank')
	   INTO prev_task_id,prev_msg
	   FROM bh_feedback
	   WHERE id=i_feedback_id;
      EXCEPTION
	 WHEN no_data_found then
	   NULL;
      END;
      
      
      -- Form the message for various status
      IF i_status_code=0 THEN
	 i_ts_msg := 'Scheduled to run';
      ELSIF i_status_code=1 THEN
	 i_ts_msg := 'Pending to run';
      ELSIF i_status_code=3 THEN
	 i_ts_msg := 'Waiting to run';
      ELSIF i_status_code=2 THEN
	 i_ts_msg := 'Job Active';
      ELSIF i_status_code=5 THEN
	 -- Get the max task id
	 SELECT max(task_id)
	   INTO i_max_task_id
	   FROM task_status
	   WHERE job_id=i_job_id;
	   
	 BEGIN
            -- i_ts_msg is defined as max 500 characters
	    SELECT  'Job Failed! ' ||substr(message,1,480)
	     INTO i_ts_msg
	    FROM task_status
	    WHERE job_id=i_job_id
	    AND status_code=5
	      AND task_id=i_max_task_id;
	 EXCEPTION
	    WHEN no_data_found THEN
	      dbms_output.put_line('Job Failed ');
	 END;
	 
	 DELETE FROM bh_feedback
	   WHERE id=i_feedback_id;
	 commit;
      ELSIF i_status_code=4 THEN
	 BEGIN
	    -- first check if the task is
	    -- actually the final task
	    -- in the job run
	    -- and if there are other 
	    -- tasks still in process
	    
	    SELECT 1
	      INTO task_in_process
	      FROM dual
	      WHERE EXISTS ( SELECT 1
	      FROM task_status
	      WHERE job_id=i_job_id
	      AND status_code=1000);
	    
	
	 EXCEPTION
	    WHEN no_data_found THEN
	      task_in_process := 0;
	 END;
	 
	 -- if there are no task in progress then 
	 -- it is the final task
	 IF task_in_process=0 then
	    BEGIN
	       SELECT  substr(message,1,500), status_code
	       INTO i_ts_msg, i_sc
	       FROM task_status
	       WHERE job_id=i_job_id
	       AND (status_code=4 or status_code=5)
	       AND task_id IS null;

               IF i_sc = 4 THEN
                 dbms_output.put_line('Job completed to success');
               ELSE
                 dbms_output.put_line('Job completed to failure');
               END IF;

	    EXCEPTION
	       WHEN no_data_found THEN
	         dbms_output.put_line('Job completed to success');
	    END;

	    DELETE FROM bh_feedback
	    WHERE id=i_feedback_id;
	    commit;
	 ELSE
	    i_status_code := 0;
	 END IF;
	 
      ELSE
         -- get the message from task status table
	 IF i_task_id != 0 AND i_job_id != 0 THEN
	    BEGIN
               SELECT  substr(message,1,500)
	       INTO i_ts_msg
	       FROM task_status
	       WHERE task_id=i_task_id;
            EXCEPTION
                WHEN no_data_found THEN
	           NULL;
	    END;
	 END IF;
      END IF;
      
      
      
      -- Compare the current value of task id with the previous value and output required info
      -- get currently processed busy hour name
      IF (i_task_id != prev_task_id OR i_ts_msg != prev_msg) THEN 
	 IF (substr(i_ts_msg,1,24) ='Report generator started') THEN
	    IF (prev_msg ='blank' OR prev_msg='Job Active' OR prev_msg='Scheduled to run'
	      OR prev_msg='Pending to run' OR prev_msg='Waiting to run') then
	       dbms_output.put_line(substr(i_ts_msg, 1, 255));
	    END IF;
	 ELSE
	    dbms_output.put_line(substr(i_ts_msg, 1, 255));
	 END IF;
	 
	 -- Update with the current task id
	 UPDATE bh_feedback
	   SET task_id=i_task_id,message=i_ts_msg
	   WHERE id=i_feedback_id;
	 commit;
      end IF;
      
      o_status_id := i_status_code;
      
   EXCEPTION
      WHEN deleted THEN
	dbms_output.put_line('================');
	dbms_output.put_line('Job deleted ');
	dbms_output.put_line('================');
	o_status_id := i_status_code;   
	DELETE FROM bh_feedback
	  WHERE id=i_feedback_id;  
      WHEN OTHERS THEN
	dbms_output.put_line('Error :' || SQLERRM);
   END poll;
  
   
   PROCEDURE add_to_error_string(table_name VARCHAR2) IS
   BEGIN
      
      gErrorStringSet := true;
      IF (instr(gErrorString, table_name) = 0) THEN
	 gErrorString := gErrorString || ' ' || table_name;	 
      END IF;
      
   END;

   --------------------------------------------------------------------------------
   -- Used by the ReportGenerator to transfer data from rg_pabh_tmp to           --
   -- rg_stored_pabh and keep pm_pabh availability window up to date.            --
   --------------------------------------------------------------------------------
   PROCEDURE UpdatePABH(bhid IN INTEGER) IS
      this_tab_done BOOLEAN := false;
      retry_count   INTEGER := 0;
      sleep_time    INTEGER := SLEEP_TIME_INITIAL;
      startTime     DATE;
      endTime       DATE;

      currFirst     DATE;
      currLast      DATE;
      bh_scenario_id INTEGER;      

      l_sql_del     VARCHAR2(2000);
      l_sql_ins     VARCHAR2(2000);
   BEGIN 

      SELECT max(tstamp), min(tstamp) INTO endTime, startTime
	FROM rg_pabh_tmp;
      
      SELECT scenario_id INTO bh_scenario_id
	FROM pm_busyhour
	WHERE id = bhid;
                 
        l_sql_del := 'DELETE FROM rg_stored_pabh '
                        || 'WHERE summary_id = :1 '
                        || 'AND tstamp BETWEEN :2 AND ( :3 + 3599/86400) '
                        || 'AND instance_id IN (SELECT distinct instance_id '
                        || 'FROM rg_pabh_tmp) ';

        l_sql_ins := 'INSERT INTO rg_stored_pabh '
                        || '(summary_id, instance_id, tstamp, hour, sundayValue, '
                        || 'mondayValue, tuesdayValue, wednesdayValue, thursdayValue, '
                        || 'fridayValue, saturdayValue) '
                        || 'SELECT '
                        || ':1, instance_id, tstamp, hour, sundayValue, '
                        || 'mondayValue, tuesdayValue, wednesdayValue, thursdayValue, '
                        || 'fridayValue, saturdayValue FROM RG_PABH_TMP ';

      WHILE (NOT this_tab_done) LOOP
      
	 this_tab_done := true;
	 retry_count := retry_count + 1;
	 
	 BEGIN 
	    
            EXECUTE IMMEDIATE l_sql_del USING bhid, startTime, endTime;
	    
            EXECUTE IMMEDIATE l_sql_ins USING bhid;

	      endTime := wm_date.trunc_to_last_period(endTime, 'W');

	      SELECT firstavail, lastavail INTO currFirst, currLast
     		FROM pm_pabh
     	       WHERE id = bhid;

   		IF currFirst is null or startTime < currFirst
   		THEN
       		    update pm_pabh set firstavail = startTime where id = bhid;
   		END IF;

   		IF currLast is null or startTime > currLast
   		THEN
       		    update pm_pabh set lastavail = endTime where id = bhid;
   		END IF;

	 EXCEPTION 
	    WHEN OTHERS THEN
	      this_tab_done := false;
	 END;
	 
	 this_tab_done := this_tab_done OR (retry_count > MAX_RETRY);
	 
	 IF (NOT this_tab_done) THEN
	    dbms_lock.sleep(sleep_time);
	    sleep_time := sleep_time * SLEEP_TIME_MULT;
	 END IF;
	    	 
      END LOOP;
      
      --
      -- If we exited because the retries were exhausted,
      -- add the table name to the list.
      -- 
      IF (retry_count > MAX_RETRY) THEN 
	 add_to_error_string('RG_STORED_PABH');
      END IF;

   END updatePABH;
   
   
   -- ------------------------------------------------------------------------------
   --
   -- ------------------------------------------------------------------------------
   PROCEDURE clean_temp_tables(task_id_in INTEGER) IS
   BEGIN
      
      DELETE FROM bh_refresh_sbh WHERE task_id IN
	(SELECT job.task_id FROM active_task tsk, active_task job
	WHERE tsk.task_id = task_id_in AND tsk.job_id = job.job_id);
      
      -- We truncate to be safe...
      execute immediate 'TRUNCATE TABLE bh_driver_tmp';
      
   END clean_temp_tables;


   -- ------------------------------------------------------------------------------
   -- create_bh_sessions
   -- ------------------------------------------------------------------------------
   PROCEDURE create_bh_sessions( p_jobid IN INTEGER, p_bh_sessions IN INTEGER )
   IS
      bh_taskid INTEGER;
      task_id INTEGER;
   BEGIN

      -- We need to truncate any existing session_id_tab as we can be called
      -- each time through the loop of weeks
      session_id_tab.delete;

      FOR i IN 1..p_bh_sessions LOOP

         -- Create a task for the session so that we can track the status
         task_id := Status.generic_start_task( p_jobid, 'Session: preparing to submit' );

         -- I would have liked to just use JOB as the only parameter but we lose the
         -- ability of the job to update its own task
         dbms_job.submit( bh_taskid, 'bh_update.bh_session( JOB, '
                     || to_char(p_jobid) 
                     || ', '
                     || to_char(task_id)
                     || ' );' );

         -- Update the global table for session ids
         session_id_tab.extend;
         session_id_tab( session_id_tab.last ) := bh_session_id_rowtype( bh_taskid );

         Status.update_status( p_jobid, 
                   task_id, 
                   Sched_general.TM_Message,
                   'Session: ' || to_char(bh_taskid) );
      END LOOP;

      EXCEPTION
         WHEN OTHERS THEN
            RAISE;

   END create_bh_sessions;

   -- -----------------------------------------------------------------------
   -- Have all of the sessions completed?
   -- -----------------------------------------------------------------------
   FUNCTION all_sessions_complete RETURN BOOLEAN
   IS
      l_status BOOLEAN := FALSE;

      -- Use our global session id table
      CURSOR job_cur IS
         SELECT 1
         FROM user_jobs uj, TABLE( CAST( session_id_tab as bh_session_id_tabletype ) ) s
         WHERE s.job_id = uj.job;

      job_rec job_cur%ROWTYPE;

   BEGIN

      -- We are looking to see if anything is still processing
      OPEN job_cur;
      FETCH job_cur INTO job_rec;
      l_status := job_cur%NOTFOUND;
      CLOSE job_cur;

      RETURN l_status;

   EXCEPTION
      WHEN OTHERS THEN

         IF ( job_cur%ISOPEN ) THEN
            CLOSE job_cur;
         END IF;

         RAISE;

   END all_sessions_complete;

   -- -----------------------------------------------------------------------
   -- Are there READY tasks waiting?
   -- -----------------------------------------------------------------------
   FUNCTION waiting_tasks( p_jobid IN INTEGER ) RETURN BOOLEAN
   IS
      l_status BOOLEAN := FALSE;

      CURSOR task_cur (c_job_id IN INTEGER, c_code IN INTEGER) IS
         SELECT 1
         FROM busyhour_task
         WHERE job_id = c_job_id
         AND status_code = c_code;

      task_rec task_cur%ROWTYPE;
   BEGIN

      -- We are looking to see if any tasks are marked as READY
      OPEN task_cur( p_jobid, g_OPS_READY );
      FETCH task_cur INTO task_rec;
      l_status := task_cur%FOUND;
      CLOSE task_cur;

      RETURN l_status;

   EXCEPTION
      WHEN OTHERS THEN

         IF ( task_cur%ISOPEN ) THEN
            CLOSE task_cur;
         END IF;

         RAISE;

   END waiting_tasks;

   -- -----------------------------------------------------------------------
   -- This is the SESSION for busyhour. 
   -- Once started it continues to run as long as it receives new tasks.
   -- -----------------------------------------------------------------------
   PROCEDURE bh_session( p_orajobid IN INTEGER, p_jobid IN INTEGER, p_taskid IN INTEGER)
   IS
      success BOOLEAN := FALSE;
      needs_task BOOLEAN := TRUE;
      l_task BHtaskRec;
      l_override VARCHAR2(2);

      vprc VARCHAR2(30) := 'bh_session';
   BEGIN

      -- Let them know we are starting
      dbms_application_info.set_module( g_module_name, vprc );

      Status.update_status(p_jobid, p_taskid, sched_general.TM_Message,
         'Session ' || to_char(p_orajobid) || ' started' );

      -- init the session and alter the tunables
      AlterTunables;

      BEGIN
        SELECT nvl(override,'N') into l_override
        from active_job aj,bh_parameter bp
        where job_id = p_jobid and
          sched_fk = bp.id;
        EXCEPTION
          WHEN NO_DATA_FOUND then
            dbms_output.put_line('No data found for the override flag for this job in bh session.');
            l_override := 'N';
      END;

      gOverRide := l_override;

      WHILE ( needs_task ) LOOP

         -- Get the next task
         get_next_task(p_jobid, l_task);

         IF (l_task.task_id IS NULL ) THEN
            -- No more work available
            needs_task := FALSE;
         ELSE
            IF ( SummarizeTable( l_task ) ) THEN
               update_bh_task( l_task.job_id, l_task.task_id, g_OPS_SUCCESS );
            ELSE
               -- We failed but do we retry?
               IF ( l_task.retry_count < MAX_RETRY ) THEN
                  -- RESET for another pass, READY and retry_count++


                  update_bh_task( l_task.job_id, l_task.task_id, g_OPS_READY, TRUE );
               ELSE
                  update_bh_task( l_task.job_id, l_task.task_id, g_OPS_FAILED );
               END IF;
            END IF;

            COMMIT;

         END IF;

      END LOOP;

      -- finalize and quit
      Status.generic_end_task( p_jobid, p_taskid, 
         'Session ' || to_char(p_orajobid) || ' completed' );

      -- Finally
      COMMIT;

   END bh_session;

   -- -----------------------------------------------------------------------
   -- Get the next READY task
   -- -----------------------------------------------------------------------
   PROCEDURE get_next_task( p_jobid IN INTEGER, 
                            p_task IN OUT NOCOPY BHtaskRec )
   IS

      -- Force the retries to the end of the list
      CURSOR task_cur (c_job_id IN INTEGER, c_code IN INTEGER) IS
         SELECT *
         FROM busyhour_task
         WHERE job_id = c_job_id
         AND status_code = c_code
         ORDER BY retry_count, task_id
         FOR UPDATE;

      task_rec task_cur%ROWTYPE;

      needs_task  BOOLEAN := TRUE;

      retry_count INTEGER := 0;
      max_retries INTEGER := 5;   -- This is not configurable, only the g_retry_interval

   BEGIN
      -- block until we have exclusive access...
      lock_util.get_x(g_task_lock_name);

      -- NULL the record
      p_task := NULL;

      -- There are 3 possible outcomes
      --   We acquire a task
      --   There are no more tasks available
      --   We fail to acquire a task due to a rule so we retry ... for how long?

      WHILE ( needs_task ) LOOP


         IF ( retry_count > 0 ) THEN
            dbms_lock.sleep( g_retry_interval );
         END IF;

         -- Try a single fetch ... hopefully we get a task
         OPEN task_cur( p_jobid, g_OPS_READY );
         FETCH task_cur INTO task_rec;

         IF ( task_cur%FOUND ) THEN

            UPDATE busyhour_task
               SET status_code = g_OPS_ACTIVE
               WHERE CURRENT OF task_cur;

            -- Build p_taskRec for return
            p_task.job_id := task_rec.job_id;
            p_task.task_id := task_rec.task_id;
            p_task.status_code := task_rec.status_code;
            p_task.start_date := task_rec.start_date;
            p_task.end_date := task_rec.end_date;
            p_task.retry_count := task_rec.retry_count;
            p_task.table_id := task_rec.table_id;
            p_task.table_name := task_rec.table_name;
            p_task.entity_id := task_rec.entity_id;
            p_task.is_time_series := task_rec.is_time_series;
            p_task.summary_id := task_rec.summary_id;
            p_task.start_id := task_rec.start_id;
            p_task.end_id := task_rec.end_id;

            CLOSE task_cur;
            needs_task := FALSE;

            COMMIT;

         -- We didn't find a task ... but why?
         ELSE
   
            CLOSE task_cur;

            -- Are there any tasks in the READY state?
            IF ( waiting_tasks( p_jobid ) ) THEN

               -- We must have hit a rule limitation so we retry ... later
               retry_count := retry_count + 1;

               -- Is is time to stop the looping?
               IF ( retry_count >= max_retries ) THEN
                  needs_task := FALSE;
               END IF;

            ELSE
               -- No tasks are available
               needs_task := FALSE;
            END IF;
         END IF;

      END LOOP;

      lock_util.release(g_task_lock_name);

      EXCEPTION
         WHEN OTHERS THEN
            IF ( task_cur%ISOPEN ) THEN
               CLOSE task_cur;
            END IF;

            -- PL/SQL needs a "finally" block :-)
            lock_util.release(g_task_lock_name);

            RAISE;

   END get_next_task;

   -- -----------------------------------------------------------------------
   -- test
   -- -----------------------------------------------------------------------
   PROCEDURE task_poll_x( p_jobid IN INTEGER )
   IS
   BEGIN
      IF ( task_poll( p_jobid ) ) THEN
         dbms_output.put_line('DONE');
      ELSE
         dbms_output.put_line('ERROR');
      END IF;
   END task_poll_x;

   -- -----------------------------------------------------------------------
   -- test
   -- -----------------------------------------------------------------------
   PROCEDURE get_next_task_x( p_jobid IN INTEGER )
   IS
      x BHtaskRec;
   BEGIN
      get_next_task(p_jobid, x);

      IF ( x.task_id IS NULL ) THEN
         dbms_output.put_line('DONE');
      ELSE
         dbms_output.put_line(x.task_id);
      END IF;
   END get_next_task_x;

   -- -----------------------------------------------------------------------
   -- Update the busyhour_task
   -- -----------------------------------------------------------------------
   PROCEDURE update_bh_task( p_jobid IN INTEGER, 
                             p_taskid IN INTEGER, 
                             p_code IN INTEGER,
                             p_reset IN BOOLEAN DEFAULT FALSE)
   IS
   BEGIN

      -- block until we have exclusive access...
      lock_util.get_x(g_task_lock_name);

      IF ( p_reset ) THEN

         -- Reset to READY and bump the retry_count
         UPDATE busyhour_task
            SET status_code = p_code,
                retry_count = retry_count + 1
         WHERE job_id = p_jobid
         AND task_id = p_taskid;

      ELSE

         UPDATE busyhour_task
            SET status_code = p_code
         WHERE job_id = p_jobid
         AND task_id = p_taskid;

      END IF;

      lock_util.release(g_task_lock_name);

   EXCEPTION
      WHEN OTHERS THEN
         lock_util.release(g_task_lock_name);
         RAISE;

   END update_bh_task;

   -- -----------------------------------------------------------------------
   -- Show how many tasks are remaining
   -- -----------------------------------------------------------------------
   PROCEDURE update_ready_count( p_jobid IN INTEGER )
   IS
      ready_count INTEGER := 0;
   BEGIN

      SELECT COUNT(*) INTO ready_count
      FROM busyhour_task
      WHERE job_id = p_jobid
      AND status_code = g_OPS_READY;

      -- We have multiple passes that may need tracking
      dbms_application_info.set_client_info( 'Tasks remaining: ' || to_char( ready_count) );

   END update_ready_count;

   -- -----------------------------------------------------------------------
   -- Poll while the sessions are running
   -- -----------------------------------------------------------------------
   FUNCTION task_poll( jobidparam IN INTEGER) RETURN BOOLEAN
   IS
      success BOOLEAN := TRUE;

      poll_jobs BOOLEAN := TRUE;

   BEGIN

      while ( poll_jobs ) LOOP

         update_ready_count( jobidparam );

         dbms_lock.sleep( g_poll_interval );

         -- Have we been interrupted or have the task completed?
         IF ( job_interrupted( jobidparam ) ) THEN

            -- Mark all READY TASKS as ABORT so that we starve the sessions
            abort_tasks( jobidparam );

	    Status.update_job_status(jobidparam, Sched_General.tm_message,
	       'BusyHour failed - interrupted');

            success := FALSE;

            -- stop the loop
            poll_jobs := FALSE;

            -- We must commit so that the tasks reflect the ABORT
            -- and the Job status is is updated
            COMMIT;

         ELSE

            -- Are we complete?
            IF ( all_sessions_complete ) THEN

               -- stop the loop
               poll_jobs := FALSE;

               -- Did all of the tasks complete with a SUCCESS?
               IF ( all_tasks_success( jobidparam ) ) THEN

	          Status.update_job_status(jobidparam, Sched_General.tm_message,
	             'BusyHour for week complete');
 
                  success := TRUE;

               ELSE

	          Status.update_job_status(jobidparam, Sched_General.tm_message,
	             'BusyHour failed');

                  success := FALSE;

               END IF;   -- all ops complete

            END IF;   -- all session complete

         END IF;   -- job interrupted

      END LOOP;

      RETURN success;

   END task_poll;

   -- -----------------------------------------------------------------------
   -- Create the busy_task record
   --     This is a child record of the active_task record.
   -- -----------------------------------------------------------------------
   PROCEDURE add_bh_task( p_jobid IN INTEGER, 
                          p_task_id IN INTEGER, 
                          p_task IN BHtaskRec )
   IS
   BEGIN

      INSERT
         INTO busyhour_task (
         job_id,
         task_id,
         status_code,
         start_date,
         end_date,
         retry_count,
         table_id,
         table_name,
         entity_id,
         is_time_series,
         summary_id,
         start_id,
         end_id
         )
         VALUES (
         p_jobid,
         p_task_id,
         p_task.status_code,
         p_task.start_date,
         p_task.end_date,
         p_task.retry_count,
         p_task.table_Id,
         p_task.table_name,
         p_task.entity_id,
         p_task.is_time_series,
         p_task.summary_id,
         p_task.start_id,
         p_task.end_id
	 );

   END add_bh_task;

   -- -----------------------------------------------------------------------
   -- Have all of the tasks completed with a SUCCESS?
   -- -----------------------------------------------------------------------
   FUNCTION all_tasks_success( p_jobid IN INTEGER ) RETURN BOOLEAN
   IS
      l_status BOOLEAN := FALSE;

      CURSOR task_cur (c_job_id IN INTEGER, c_code IN INTEGER) IS
         SELECT 1
         FROM busyhour_task
         WHERE job_id = c_job_id
         AND status_code != c_code;

      task_rec task_cur%ROWTYPE;
   BEGIN

      -- We are looking to see if any tasks are NOT marked a SUCCESS
      OPEN task_cur( p_jobid, g_OPS_SUCCESS );
      FETCH task_cur INTO task_rec;
      l_status := task_cur%NOTFOUND;
      CLOSE task_cur;

      RETURN l_status;

   EXCEPTION
      WHEN OTHERS THEN

         IF ( task_cur%ISOPEN ) THEN
            CLOSE task_cur;
         END IF;

         RAISE;

   END all_tasks_success;

   -- -----------------------------------------------------------------------
   -- Update all READY tasks to ABORT. We are trying to starve
   -- the get_next_task routine so that the sessions will complete.
   -- -----------------------------------------------------------------------
   PROCEDURE abort_tasks( p_jobid IN INTEGER )
   IS
   BEGIN

      -- block until we have exclusive access...
      lock_util.get_x(g_task_lock_name);

      UPDATE busyhour_task
         SET status_code = g_OPS_ABORT
      WHERE
         job_id = p_jobid
         AND status_code = g_OPS_READY;

      lock_util.release(g_task_lock_name);

   EXCEPTION
      WHEN OTHERS THEN
         lock_util.release(g_task_lock_name);
         RAISE;

   END abort_tasks;


   -- ------------------------------------------------------------------------------
   -- slice_msg
   -- ------------------------------------------------------------------------------
   FUNCTION slice_msg( start_id IN INTEGER, end_id IN INTEGER ) RETURN VARCHAR2
   IS
      -- NULL
      return_msg VARCHAR2(100);
   BEGIN

      IF ( start_id IS NOT NULL AND end_id IS NOT NULL ) THEN
         return_msg := ' ('
            || TO_CHAR(start_id)
            || '-'
            || TO_CHAR(end_id)
            || ') ';
      END IF;

      RETURN return_msg;

   END slice_msg;



   -- ------------------------------------------------------------------------------
   -- dump_options
   -- ------------------------------------------------------------------------------
   PROCEDURE dump_options( job_id_in IN INTEGER )
   IS
      l_task_id INTEGER;
   BEGIN

      -- Need a better interface for creating single informational status messages

      -- The defaults are hard-coded s be careful ...

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHHashAreaSize[2*g_sort_area_size] := ' || TO_CHAR(g_hash_area_size) );

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHPollInterval[10] := ' || TO_CHAR(g_poll_interval) );

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHRetryInterval[5] := ' || TO_CHAR(g_retry_interval) );

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHSessions[1] := ' || TO_CHAR(g_bh_sessions) );

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHSlices[BHSessions] := ' || TO_CHAR(g_slices) );

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHSliceThreshold[0] := ' || TO_CHAR(g_slicethreshold) );

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHSortAreaRetainedSize[TrafficSortAreaRetainedSize] := ' || TO_CHAR(g_sort_area_retained_size) );

      l_task_id := Status.generic_start_task( job_id_in, NULL);
      Status.generic_end_task( job_id_in, l_task_id,
         'CONFIG: BHSortAreaSize[TrafficSortAreaSize] := ' || TO_CHAR(g_sort_area_size) );

   END dump_options;

   -- ------------------------------------------------------------------------------
   -- job_interrupted
   -- ------------------------------------------------------------------------------
   FUNCTION job_interrupted(job_id_in IN INTEGER) RETURN BOOLEAN IS
      ret_val BOOLEAN := FALSE;
      start_time DATE;
      status_code_out INTEGER;
      taskid INTEGER;
      session_status sys.v_$session.Status%TYPE;
   BEGIN

      BEGIN

	 SELECT ts.status_code, ts.started, s.Status
	   INTO status_code_out, start_time, session_status
	   FROM task_status ts, sys.v_$session s
	   WHERE ts.job_id = job_id_in
	   AND ts.task_id IS NULL
	   AND s.audsid = USERENV('SESSIONID');

         -- Differentiate between interruptions ... and drop a message
         IF ( Sched_General.is_end_of_resource(status_code_out) ) THEN
            taskid := Status.generic_start_task(job_id_in, 'JOB INTERRUPTED: End Of Resource.');
            ret_val := TRUE;
         ELSIF ( session_status = 'KILLED' ) THEN
            taskid := Status.generic_start_task(job_id_in, 'JOB INTERRUPTED: Session was KILLED.');
            ret_val := TRUE;
         END IF;

      EXCEPTION
	 WHEN OTHERS THEN
	   ret_val := FALSE;
      END;

      RETURN ret_val;

   END job_interrupted;


END BH_Update;
/
show errors PACKAGE BODY BH_Update;
